structure Imp_rewrite :> Imp_rewrite =
struct

type hol_type = Type.hol_type;
type term = Term.term
type thm = Thm.thm
type 'a net = 'a Net.net
type goal = Abbrev.goal
type conv = Abbrev.conv
type tactic = Abbrev.tactic;

open Exception;
open Lib;
open Thm;
open Dsyntax;
open Type;
open Term;
open Type;
open Net;
open Drule;
open Tactic;
   infix THEN;
open Conv;
   infix THENC;
open boolTheory;

fun IMP_REWR_ERR{function,message} = 
         HOL_ERR{origin_structure="abstract_theory",
		 origin_function=function,
		 message=message}

local fun COUNT_UNDISCH 0 thm = thm
	| COUNT_UNDISCH n thm = COUNT_UNDISCH (n -1) (UNDISCH thm)
in
fun INST_ALL_TY_TERM (tm_subst,type_subst) th =
    let val num_hyp = length (hyp th)
    in
       COUNT_UNDISCH num_hyp(INST tm_subst(INST_TYPE type_subst(DISCH_ALL th)))
    end
end;
    
    
(* Match a given part of "th" to a term, instantiating "th".
   The part should be free in the theorem, except for outer bound variables.
*)
fun PART_MATCH_ALL partfn th =
   let val pth = GSPEC (GEN_ALL th)
       val pat = partfn(concl pth)
       val matchfn = match_term pat 
   in
   fn tm => INST_ALL_TY_TERM (matchfn tm) pth
   end;

local exception NOT_AN_EQ
in
fun IMP_REWR_CONV th =
   let val instth = PART_MATCH_ALL lhs th 
                    handle _ => raise NOT_AN_EQ
   in
   fn asl => fn tm => 
     let val eqn = instth tm
         val l = lhs(concl eqn)
         val th' = 
            if (l = tm) 
            then eqn 
            else TRANS (ALPHA tm l) eqn
                 handle _
                 => raise IMP_REWR_ERR{function = "IMP_REWR_CONV",
                                 message = "lhs of theorem doesn't match term"}
     in
     if (all (fn t => mem t asl)(hyp th'))
     then th'
     else raise IMP_REWR_ERR{function = "IMP_REWR_CONV",
                             message = "introduces new assumptions"}
     end		    
   end
   handle NOT_AN_EQ => raise IMP_REWR_ERR{function = "IMP_REWR_CONV",
                            message = "bad theorem argument (not an equation)"}
        | (e as HOL_ERR _) => raise e
end;

fun mk_imp_rewrites th =
   let val th = Drule.SPEC_ALL th
       val t = Thm.concl th 
   in 
   if (is_imp t)
   then mk_imp_rewrites(UNDISCH th)
   else
   if (is_eq t)
   then [th]
   else if (is_conj t)
        then let val (c1,c2) = Drule.CONJ_PAIR th
             in
             ((mk_imp_rewrites c1)@(mk_imp_rewrites c2))
             end
        else if (is_neg t)
             then [Drule.EQF_INTRO th]
             else [Drule.EQT_INTRO th]
   end
   handle _ => raise IMP_REWR_ERR{function = "mk_imp_rewrites",message = ""};

(* [th1; ... ; thn]  --> (mk_rewrites th1) @ ... @ (mk_rewrites thn)   *)
fun mk_imp_rewritesl thl = itlist (append o mk_imp_rewrites) thl [];

(* Augment a conversion net with a list of equality theorems *)
fun add_to_imp_net thl net =
   itlist enter
          (map (fn th => (lhs(concl th), IMP_REWR_CONV th))
               (mk_imp_rewritesl thl))
          net;

fun mk_imp_conv_net thl = add_to_imp_net thl empty_net;

(* Create a conversion from a net *)
fun IMP_REWRITES_CONV asl net tm 
    = FIRST_CONV(map (fn f=>f asl)(lookup tm net)) tm;

    
(* List of basic rewrites (made assignable to enable us to add extra
 * rewrite rules later e.g., those for the theory of pairs).
*)
val basic_imp_rewrites = ref 
   (mk_imp_conv_net (mk_imp_rewritesl [REFL_CLAUSE,
				       EQ_CLAUSES,
				       NOT_CLAUSES,
				       AND_CLAUSES,
				       OR_CLAUSES,
				       IMP_CLAUSES,
				       COND_CLAUSES,
				       FORALL_SIMP,
				       EXISTS_SIMP,
				       ABS_SIMP]));

(* =====================================================================*)
(* Main rewriting conversion                         			*)
(* =====================================================================*)

fun GEN_IMP_REWRITE_CONV (rw_func:conv->conv) basic_net thl asl = 
   rw_func (IMP_REWRITES_CONV asl (add_to_imp_net (mk_imp_rewritesl thl) 
                                                  (!basic_net)));

(* ---------------------------------------------------------------------*)
(* Rewriting conversions.                        			*)
(* ---------------------------------------------------------------------*)

val PURE_IMP_REWRITE_CONV = GEN_IMP_REWRITE_CONV Conv.TOP_DEPTH_CONV
                                                 (ref empty_net)
and IMP_REWRITE_CONV = GEN_IMP_REWRITE_CONV Conv.TOP_DEPTH_CONV 
                                            basic_imp_rewrites
and PURE_ONCE_IMP_REWRITE_CONV = GEN_IMP_REWRITE_CONV Conv.ONCE_DEPTH_CONV
                                                      (ref empty_net)
and ONCE_IMP_REWRITE_CONV = GEN_IMP_REWRITE_CONV Conv.ONCE_DEPTH_CONV
                                                 basic_imp_rewrites;

(* Main rewriting rule *)
fun GEN_IMP_REWRITE_RULE f n thl th 
    = Conv.CONV_RULE(GEN_IMP_REWRITE_CONV f n thl (hyp th))th;

val PURE_IMP_REWRITE_RULE = GEN_IMP_REWRITE_RULE Conv.TOP_DEPTH_CONV
                                                 (ref empty_net)
and IMP_REWRITE_RULE = GEN_IMP_REWRITE_RULE Conv.TOP_DEPTH_CONV
                                            basic_imp_rewrites
and PURE_ONCE_IMP_REWRITE_RULE = GEN_IMP_REWRITE_RULE Conv.ONCE_DEPTH_CONV 
                                                      (ref empty_net)
and ONCE_IMP_REWRITE_RULE = GEN_IMP_REWRITE_RULE Conv.ONCE_DEPTH_CONV 
                                                 basic_imp_rewrites;

(* Rewrite a theorem with the help of its assumptions *)

fun PURE_ASM_IMP_REWRITE_RULE thl th =
   PURE_IMP_REWRITE_RULE ((map Thm.ASSUME (Thm.hyp th)) @ thl) th
and ASM_IMP_REWRITE_RULE thl th = 
   IMP_REWRITE_RULE ((map Thm.ASSUME (Thm.hyp th)) @ thl) th
and PURE_ONCE_ASM_IMP_REWRITE_RULE thl th =
   PURE_ONCE_IMP_REWRITE_RULE ((map Thm.ASSUME (Thm.hyp th)) @ thl) th
and ONCE_ASM_IMP_REWRITE_RULE thl th = 
   ONCE_IMP_REWRITE_RULE ((map Thm.ASSUME (Thm.hyp th)) @ thl) th;



fun GEN_IMP_REWRITE_TAC f n thl (g:goal as (asl,t)) 
    = Conv.CONV_TAC(GEN_IMP_REWRITE_CONV f n thl asl)g;
      

val (PURE_IMP_REWRITE_TAC: Thm.thm list -> tactic) =
   GEN_IMP_REWRITE_TAC Conv.TOP_DEPTH_CONV (ref empty_net)
and (IMP_REWRITE_TAC: Thm.thm list -> tactic) =
   GEN_IMP_REWRITE_TAC Conv.TOP_DEPTH_CONV basic_imp_rewrites
and (PURE_ONCE_IMP_REWRITE_TAC: Thm.thm list -> tactic) =
   GEN_IMP_REWRITE_TAC Conv.ONCE_DEPTH_CONV (ref empty_net)
and (ONCE_IMP_REWRITE_TAC: Thm.thm list -> tactic) = 
   GEN_IMP_REWRITE_TAC Conv.ONCE_DEPTH_CONV basic_imp_rewrites;


(* Rewrite a goal with the help of its assumptions *)

fun PURE_ASM_IMP_REWRITE_TAC thl :tactic = 
   Tactical.ASSUM_LIST (fn asl => PURE_IMP_REWRITE_TAC (asl @ thl))
and ASM_IMP_REWRITE_TAC thl :tactic      = 
   Tactical.ASSUM_LIST (fn asl => IMP_REWRITE_TAC (asl @ thl))
and PURE_ONCE_ASM_IMP_REWRITE_TAC thl :tactic = 
   Tactical.ASSUM_LIST (fn asl => PURE_ONCE_IMP_REWRITE_TAC (asl @ thl))
and ONCE_ASM_IMP_REWRITE_TAC thl :tactic      = 
   Tactical.ASSUM_LIST (fn asl => ONCE_IMP_REWRITE_TAC (asl @ thl));


(* Rewriting using equations that satisfy a predicate  *)
fun FILTER_PURE_ASM_IMP_REWRITE_RULE f thl th =
    PURE_IMP_REWRITE_RULE ((map Thm.ASSUME (filter f (Thm.hyp th))) @ thl) th
and FILTER_ASM_IMP_REWRITE_RULE f thl th = 
    IMP_REWRITE_RULE ((map Thm.ASSUME (filter f (Thm.hyp th))) @ thl) th
and FILTER_PURE_ONCE_ASM_IMP_REWRITE_RULE f thl th =
    PURE_ONCE_IMP_REWRITE_RULE((map Thm.ASSUME (filter f (Thm.hyp th)))@thl) th
and FILTER_ONCE_ASM_IMP_REWRITE_RULE f thl th = 
    ONCE_IMP_REWRITE_RULE ((map Thm.ASSUME (filter f (Thm.hyp th))) @ thl) th;;

fun FILTER_PURE_ASM_IMP_REWRITE_TAC f thl =
    Tactical.ASSUM_LIST 
          (fn asl => PURE_IMP_REWRITE_TAC ((filter (f o Thm.concl) asl)@thl))
and FILTER_ASM_IMP_REWRITE_TAC f thl =
    Tactical.ASSUM_LIST
          (fn asl => IMP_REWRITE_TAC ((filter (f o Thm.concl) asl) @ thl))
and FILTER_PURE_ONCE_ASM_IMP_REWRITE_TAC f thl =
    Tactical.ASSUM_LIST
     (fn asl => PURE_ONCE_IMP_REWRITE_TAC ((filter (f o Thm.concl) asl) @ thl))
and FILTER_ONCE_ASM_IMP_REWRITE_TAC f thl =
    Tactical.ASSUM_LIST 
         (fn asl => ONCE_IMP_REWRITE_TAC ((filter (f o Thm.concl) asl) @ thl));

end;
