(* port to HOL98 of the abstract theory package by David Shepherd
   for HOL90, which itself is based on Phillip Windley's abstract
   theory package for HOL88

   Dirk Nowotka, Turku Centre for Computer Science, Finland

   July 1999
*)

signature abs_theoryLib =
sig

  type hol_type = Type.hol_type
  type term = Term.term
  type thm = Thm.thm
  type goal = Abbrev.goal
  type tactic = Abbrev.tactic
  type conv = Abbrev.conv
  type thm_tactic = Abbrev.thm_tactic

  val theory_obligs: ((string * {oblig: thm, rep_defs: thm list}) list) ref

  val new_abstract_entity: {Name: string, Predicate: term} -> thm

  val abs_convert_term: term -> term

  val ag:            term frag list -> GoalstackPure.proofs
  val abs_set_goal:  goal -> GoalstackPure.proofs
  val abs_prove:     term * tactic -> thm
  val abs_store_thm: string * term * tactic -> thm

  val EXPAND_THOBS_CONV: conv
  val EXPAND_THOBS_TAC:  tactic
  val EXPAND_THOBS_RULE: thm -> thm
  val STRIP_THOBS_THEN:  thm_tactic -> thm_tactic -> tactic
  val STRIP_THOBS_TAC:   tactic

  val get_obligs:            string -> thm list
  val get_abs_reps:          string -> thm list
  val add_obligs:            string -> unit

  val instantiate_abs_thm: {abs_term: term,
                            rep_term: term,
                            validation: thm} list
                           -> thm -> thm
  val instantiate_abstract_definition: {abs_term: term, rep_term: term} list
                                       -> thm -> thm -> thm

  val mk_imp_conv_net : thm list -> (term list -> conv) Net.net
  val IMP_REWRITES_CONV : term list -> (term list -> conv) Net.net -> conv
  val basic_imp_rewrites : (term list -> conv) Net.net ref
  val GEN_IMP_REWRITE_CONV : (conv -> conv) -> (term list -> conv) Net.net ref
                              -> thm list -> term list -> conv
  val PURE_IMP_REWRITE_CONV : thm list -> term list -> conv
  val IMP_REWRITE_CONV : thm list -> term list -> conv
  val PURE_ONCE_IMP_REWRITE_CONV : thm list -> term list -> conv
  val ONCE_IMP_REWRITE_CONV : thm list -> term list -> conv
  val GEN_IMP_REWRITE_RULE : (conv -> conv) -> (term list -> conv) Net.net ref
                              ->  thm list -> thm -> thm
  val GEN_IMP_REWRITE_TAC : (conv -> conv) -> (term list -> conv) Net.net ref
                              -> thm list -> tactic
  val PURE_IMP_REWRITE_RULE : thm list -> thm -> thm
  val IMP_REWRITE_RULE : thm list -> thm -> thm
  val PURE_ONCE_IMP_REWRITE_RULE : thm list -> thm -> thm
  val ONCE_IMP_REWRITE_RULE : thm list -> thm -> thm
  val PURE_ASM_IMP_REWRITE_RULE : thm list -> thm -> thm
  val ASM_IMP_REWRITE_RULE : thm list -> thm -> thm
  val PURE_ONCE_ASM_IMP_REWRITE_RULE : thm list -> thm -> thm
  val ONCE_ASM_IMP_REWRITE_RULE : thm list -> thm -> thm
  val PURE_IMP_REWRITE_TAC : thm list -> tactic 
  val IMP_REWRITE_TAC : thm list -> tactic
  val PURE_ONCE_IMP_REWRITE_TAC : thm list -> tactic
  val ONCE_IMP_REWRITE_TAC : thm list -> tactic
  val PURE_ASM_IMP_REWRITE_TAC : thm list -> tactic
  val ASM_IMP_REWRITE_TAC : thm list -> tactic
  val PURE_ONCE_ASM_IMP_REWRITE_TAC : thm list -> tactic
  val ONCE_ASM_IMP_REWRITE_TAC : thm list -> tactic
  val FILTER_PURE_ASM_IMP_REWRITE_RULE :(term -> bool) -> thm list 
                                         -> thm -> thm
  val FILTER_ASM_IMP_REWRITE_RULE :(term -> bool) -> thm list -> thm -> thm
  val FILTER_PURE_ONCE_ASM_IMP_REWRITE_RULE :(term -> bool) -> thm list 
                                              -> thm -> thm
  val FILTER_ONCE_ASM_IMP_REWRITE_RULE :(term -> bool)
                                         -> thm list -> thm -> thm
  val FILTER_PURE_ASM_IMP_REWRITE_TAC :(term -> bool) -> thm list -> tactic
  val FILTER_ASM_IMP_REWRITE_TAC :(term -> bool) -> thm list -> tactic
  val FILTER_PURE_ONCE_ASM_IMP_REWRITE_TAC :(term -> bool) 
                                              -> thm list -> tactic
  val FILTER_ONCE_ASM_IMP_REWRITE_TAC :(term -> bool) -> thm list -> tactic

end;

